# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dydxprotocol/clob/clob_pair.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...amino import amino_pb2 as amino_dot_amino__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!dydxprotocol/clob/clob_pair.proto\x12\x11\x64ydxprotocol.clob\x1a\x11\x61mino/amino.proto\"[\n\x15PerpetualClobMetadata\x12\x14\n\x0cperpetual_id\x18\x01 \x01(\r:,\x8a\xe7\xb0*\'dydxprotocol/clob/PerpetualClobMetadata\"j\n\x10SpotClobMetadata\x12\x15\n\rbase_asset_id\x18\x01 \x01(\r\x12\x16\n\x0equote_asset_id\x18\x02 \x01(\r:\'\x8a\xe7\xb0*\"dydxprotocol/clob/SpotClobMetadata\"\xc6\x04\n\x08\x43lobPair\x12\n\n\x02id\x18\x01 \x01(\r\x12y\n\x17perpetual_clob_metadata\x18\x02 \x01(\x0b\x32(.dydxprotocol.clob.PerpetualClobMetadataB,\xb2\xe7\xb0*\'dydxprotocol/clob/PerpetualClobMetadataH\x00\x12j\n\x12spot_clob_metadata\x18\x03 \x01(\x0b\x32#.dydxprotocol.clob.SpotClobMetadataB\'\xb2\xe7\xb0*\"dydxprotocol/clob/SpotClobMetadataH\x00\x12\x1a\n\x12step_base_quantums\x18\x04 \x01(\x04\x12\x19\n\x11subticks_per_tick\x18\x05 \x01(\r\x12#\n\x1bquantum_conversion_exponent\x18\x06 \x01(\x11\x12\x32\n\x06status\x18\x07 \x01(\x0e\x32\".dydxprotocol.clob.ClobPair.Status\"\xaa\x01\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x11\n\rSTATUS_ACTIVE\x10\x01\x12\x11\n\rSTATUS_PAUSED\x10\x02\x12\x16\n\x12STATUS_CANCEL_ONLY\x10\x03\x12\x14\n\x10STATUS_POST_ONLY\x10\x04\x12\x17\n\x13STATUS_INITIALIZING\x10\x05\x12\x1b\n\x17STATUS_FINAL_SETTLEMENT\x10\x06\x42\n\n\x08metadataB8Z6github.com/dydxprotocol/v4-chain/protocol/x/clob/typesb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dydxprotocol.clob.clob_pair_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z6github.com/dydxprotocol/v4-chain/protocol/x/clob/types'
  _PERPETUALCLOBMETADATA._options = None
  _PERPETUALCLOBMETADATA._serialized_options = b'\212\347\260*\'dydxprotocol/clob/PerpetualClobMetadata'
  _SPOTCLOBMETADATA._options = None
  _SPOTCLOBMETADATA._serialized_options = b'\212\347\260*\"dydxprotocol/clob/SpotClobMetadata'
  _CLOBPAIR.fields_by_name['perpetual_clob_metadata']._options = None
  _CLOBPAIR.fields_by_name['perpetual_clob_metadata']._serialized_options = b'\262\347\260*\'dydxprotocol/clob/PerpetualClobMetadata'
  _CLOBPAIR.fields_by_name['spot_clob_metadata']._options = None
  _CLOBPAIR.fields_by_name['spot_clob_metadata']._serialized_options = b'\262\347\260*\"dydxprotocol/clob/SpotClobMetadata'
  _PERPETUALCLOBMETADATA._serialized_start=75
  _PERPETUALCLOBMETADATA._serialized_end=166
  _SPOTCLOBMETADATA._serialized_start=168
  _SPOTCLOBMETADATA._serialized_end=274
  _CLOBPAIR._serialized_start=277
  _CLOBPAIR._serialized_end=859
  _CLOBPAIR_STATUS._serialized_start=677
  _CLOBPAIR_STATUS._serialized_end=847
# @@protoc_insertion_point(module_scope)
